# -*- coding: utf-8 -*-
from typing import Union

import re

from collections import deque
from pathlib import Path
from warnings import warn

import ply.lex as lex
import ply.yacc as yacc

from loguru import logger

from ..body import Body
from ..material import Composition, Element, Material
from ..surface import create_surface
from ..transformation import Transformation
from ..universe import Universe, produce_universes

warn(
    "The module 'mcnp_input_parser' is deprecated. Use mckit.parser.mcnp_input_sly_parser instead.",
    DeprecationWarning,
)

__DEBUG__ = False

# lex.lex(debug=True, debuglog=log)
# yacc.yacc(debug=True, debuglog=log)

literals = ["+", "-", ":", "*", "(", ")", "#", "."]


CELL_KEYWORDS = {
    "IMP",
    "VOL",
    "PWT",
    "EXT",
    "FCL",
    "WWN",
    "DXC",
    "NONU",
    "PD",
    "TMP",
    "U",
    "TRCL",
    "LAT",
    "FILL",
    "N",
    "P",
    "E",
    "LIKE",
    "BUT",
    "RHO",
    "MAT",
}

SURFACE_TYPES = {
    "P",
    "PX",
    "PY",
    "PZ",
    "S",
    "SO",
    "SX",
    "SY",
    "SZ",
    "CX",
    "CY",
    "CZ",
    "KX",
    "KY",
    "KZ",
    "TX",
    "TY",
    "TZ",
    "C/X",
    "C/Y",
    "C/Z",
    "K/X",
    "K/Y",
    "K/Z",
    "SQ",
    "GQ",
    "X",
    "Y",
    "Z",
    "RPP",
    "BOX",
    "RCC",
}

DATA_KEYWORDS = {
    "MODE",
    "N",
    "P",
    "E",
    "VOL",
    "AREA",
    "TR",
    "IMP",
    "ESPLT",
    "TSPLT",
    "PWT",
    "EXT",
    "VECT",
    "FCL",
    "WWE",
    "WWN",
    "WWP",
    "WWG",
    "WWGE",
    "PD",
    "DXC",
    "BBREM",
    "MESH",
    "GEOM",
    "REF",
    "ORIGIN",
    "AXS",
    "VEC",
    "IMESH",
    "IINTS",
    "JMESH",
    "JINTS",
    "KMESH",
    "KINTS",
    "SDEF",
    "CEL",
    "SUR",
    "ERG",
    "TME",
    "DIR",
    "VEC",
    "NRM",
    "POS",
    "RAD",
    "EXT",
    "AXS",
    "X",
    "Y",
    "Z",
    "CCC",
    "ARA",
    "WGT",
    "EFF",
    "PAR",
    "TR",
    "SI",
    "SP",
    "SB",
    "H",
    "L",
    "A",
    "S",
    "D",
    "C",
    "V",
    "DS",
    "T",
    "Q",
    "SC",
    "SSW",
    "SYM",
    "PTY",
    "SSR",
    "OLD",
    "NEW",
    "COL",
    "PSC",
    "POA",
    "BCW",
    "KCODE",
    "KSRC",
    "F",
    "FC",
    "E",
    "T",
    "C",
    "FQ",
    "FM",
    "DE",
    "DF",
    "LOG",
    "LIN",
    "EM",
    "TM",
    "CM",
    "CF",
    "SF",
    "FS",
    "SD",
    "FU",
    "TF",
    "DD",
    "DXT",
    "FT",
    "FMESH",
    "GEOM",
    "ORIGIN",
    "AXS",
    "VEC",
    "IMESH",
    "IINTS",
    "JMESH",
    "JINTS",
    "KMESH",
    "KINTS",
    "EMESH",
    "EINTS",
    "FACTOR",
    "OUT",
    "TR",
    "M",
    "GAS",
    "ESTEP",
    "NLIB",
    "PLIB",
    "PNLIB",
    "ELIB",
    "COND",
    "MPN",
    "DRX",
    "TOTNU",
    "NONU",
    "AWTAB",
    "XS",
    "VOID",
    "PIKMT",
    "MGOPT",
    "PHYS",
    "TMP",
    "THTME",
    "MT",
    "CUT",
    "ELPT",
    "NOTRN",
    "NPS",
    "CTME",
    "PRDMP",
    "LOST",
    "DBCN",
    "FILES",
    "PRINT",
    "TALNP",
    "MPLOT",
    "PTRAC",
    "PERT",
    "RAND",
    "GEN",
    "SEED",
    "STRIDE",
    "HIST",
}

COMMON_KEYWORDS = {"R", "I", "ILOG", "J", "NO", "MESSAGE"}

KEYWORDS = CELL_KEYWORDS.union(DATA_KEYWORDS)

# List of token names
tokens = (
    [
        "blank_line",
        "line_comment",
        "card_comment",
        "continue",
        "separator",
        "surface_type",
        "int_number",
        "flt_number",
        "keyword",
        "title",
        "void_material",
        "lib_spec",
    ]
    + list(KEYWORDS)
    + list(COMMON_KEYWORDS)
)


states = (
    ("continue", "exclusive"),
    ("cells", "exclusive"),
    ("ckw", "exclusive"),
    ("surfs", "exclusive"),
    ("data", "exclusive"),
)

LINE_COMMENT = r"^[ ]{0,4}C.*"
BLANK_LINE = r"\n(?=[ ]*$)"
CARD_COMMENT = r"\$.*"
CARD_START = r"^[ ]{0,4}[^C\s]"
NEWLINE_SKIP = r"\n(?=" + LINE_COMMENT + r"|[ ]{5,}[^\s])"
RESET_CONTINUE = r"\n(?=[ ]{5,}[^\s])"
CONTINUE = r"&(?=[ ]*(" + CARD_COMMENT + r")?$)"
SEPARATOR = r"\n(?=" + CARD_START + r")"
FRACTION = r"\."
EXPONENT = r"([eE][-+]?\d+)"
INT_NUMBER = r"(\d+)"
FLT_NUMBER = (
    r"("
    + INT_NUMBER
    + r"?"
    + FRACTION
    + INT_NUMBER
    + EXPONENT
    + r"?|"
    + INT_NUMBER
    + FRACTION
    + r"?"
    + EXPONENT
    + r"|"
    + INT_NUMBER
    + FRACTION
    + r")(?=[ \n-+])"
)
# TODO dvp: why FLT_NUMBER is not just: [-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?
LIB_SPEC = INT_NUMBER + r"[CDEPUY]"
KEYWORD = r"[A-Z]+(/[A-Z]+)?"
VOID_MATERIAL = r" 0 "
SKIP = r"[=, ]"

t_ANY_ignore = SKIP

_card_comments = deque()
_line_comments = []


def t_title(t):
    r""".+"""
    t.lexer.section_index = 0
    mcnp_input_lexer.lineno = 1
    t.lineno = 1
    t.lexer.last_pos = 1
    t.lexer.begin("cells")
    # t.mcnp_input_lexer.push_state('continue')
    _line_comments.clear()
    _card_comments.clear()
    return t


@lex.TOKEN(BLANK_LINE)
def t_cells_ckw_surfs_data_blank_line(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    t.lexer.section_index += 1
    if t.lexer.section_index == 1:
        t.lexer.begin("surfs")
    else:
        t.lexer.begin("data")
    t.lexer.push_state("continue")
    if _card_comments and _card_comments[-1] is not None:
        _card_comments.append(None)
    return t


@lex.TOKEN(LINE_COMMENT)
def t_continue_cells_ckw_surfs_data_line_comment(t):
    _line_comments.append(t)


@lex.TOKEN(CARD_COMMENT)
def t_continue_cells_ckw_surfs_data_card_comment(t):
    _card_comments.append(t)  # return t


def t_ANY_error(t):
    column = t.lexer.lexpos - t.lexer.last_pos + 1
    msg = r"Illegal character '{0}' at line {1} column {2}".format(
        t.value[0], t.lexer.lineno, column
    )
    raise ValueError(msg, t.value[0], t.lexer.lineno, column)


@lex.TOKEN(CONTINUE)
def t_cells_ckw_surfs_data_continue(t):
    t.lexer.push_state("continue")


@lex.TOKEN(RESET_CONTINUE)
def t_continue_reset_continue(t):
    t.lexer.pop_state()
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos


@lex.TOKEN(SEPARATOR)
def t_continue_separator(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    t.lexer.pop_state()


@lex.TOKEN(SEPARATOR)
def t_ckw_separator(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    t.lexer.pop_state()
    return t


@lex.TOKEN(SEPARATOR)
def t_INITIAL_surfs_cells_data_separator(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    if _card_comments and _card_comments[-1] is not None:
        _card_comments.append(None)
    return t


@lex.TOKEN(FLT_NUMBER)
def t_cells_ckw_surfs_data_flt_number(t):
    t.value = float(t.value)
    return t


@lex.TOKEN(VOID_MATERIAL)
def t_cells_void_material(t):
    return t


@lex.TOKEN(KEYWORD)
def t_cells_ckw_keyword(t):
    value = t.value.upper()
    if value not in CELL_KEYWORDS:
        column = t.lexer.lexpos - t.lexer.last_pos + 1 - len(value)
        msg = r"Unknown keyword '{0}' at line {1} column {2}".format(
            t.value[0], t.lexer.lineno, column
        )
        raise ValueError(msg, t.value[0], t.lexer.lineno, column)
    t.type = value
    t.value = value
    if t.lexer.current_state() == "cells":
        t.lexer.push_state("ckw")
    return t


@lex.TOKEN(KEYWORD)
def t_surfs_keyword(t):
    value = t.value.upper()
    if value not in SURFACE_TYPES:
        column = t.lexer.lexpos - t.lexer.last_pos + 1 - len(value)
        msg = r"Unknown surface type '{0}' at line {1} column {2}".format(
            t.value[0], t.lexer.lineno, column
        )
        raise ValueError(msg, t.value[0], t.lexer.lineno, column)
    t.type = "surface_type"
    t.value = value
    return t


@lex.TOKEN(LIB_SPEC)
def t_data_lib_spec(t):
    t.value = t.value.upper()
    return t


@lex.TOKEN(INT_NUMBER)
def t_cells_ckw_surfs_data_int_number(t):
    t.value = int(t.value)
    return t


@lex.TOKEN(KEYWORD)
def t_data_keyword(t):
    value = t.value.upper()
    if value in KEYWORDS:
        t.type = value
        t.value = value
    else:
        column = t.lexer.lexpos - t.lexer.last_pos + 1 - len(value)
        msg = r"Unknown keyword '{0}' at line {1} column {2}".format(
            t.value[0], t.lexer.lineno, column
        )
        raise ValueError(msg, t.value[0], t.lexer.lineno, column)
    return t


@lex.TOKEN(NEWLINE_SKIP)
def t_continue_cells_ckw_surfs_data_newline_skip(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos


mcnp_input_lexer = lex.lex(
    reflags=re.MULTILINE + re.IGNORECASE + re.VERBOSE,
    debug=__DEBUG__,
    debuglog=logger if __DEBUG__ else None,
)


def extract_comments(line):
    comment = []
    while _card_comments and (
        _card_comments[0] is None or _card_comments[0].lineno < line
    ):
        _card_comments.popleft()
    while _card_comments and _card_comments[0] is not None:
        comment.append(_card_comments.popleft().value.lstrip("$ "))
    return comment


def p_model(p):
    """model_body : title separator cell_cards blank_line \
                    surface_cards blank_line \
                    data_cards blank_line
    """
    p[0] = p[1], p[3], p[5], p[7]


def p_model_without_data(p):
    """model_body : title separator cell_cards blank_line surface_cards blank_line"""
    p[0] = p[1], p[3], p[5], None


def p_cell_cards(p):
    """cell_cards : cell_cards separator cell_card
    | cell_card
    """
    if len(p) == 2:
        p[0] = {p[1]["name"]: p[1]}
    elif len(p) == 4:
        name = p[3]["name"]
        if name in p[1].keys():
            raise ValueError("Duplicate cell name {0}.".format(name))
        p[1][name] = p[3]
        p[0] = p[1]


def p_cell_card(p):
    """cell_card : int_number cell_material expression cell_options
    | int_number cell_material expression
    | int_number LIKE int_number BUT cell_options
    """
    if len(p) == 6:
        params = p[5]
        params["reference"] = p[3]
    else:
        params = p[4] if (len(p) == 5) else {}
        params["geometry"] = p[3]
        if p[2] is not None:
            params["MAT"] = {"composition": p[2][0]}
            if p[2][1] > 0:
                params["MAT"]["concentration"] = p[2][1] * 1.0e24
            else:
                params["MAT"]["density"] = abs(p[2][1])
    params["name"] = p[1]
    comment = extract_comments(p.lineno(1))
    if comment:
        params["comment"] = comment
    p[0] = params


def p_cell_options(p):
    """cell_options : cell_options cell_option
    | cell_option"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[1].update(p[2])
        p[0] = p[1]


def p_cell_option(p):
    """cell_option : fill_option
    | trcl_option
    | cell_float_option
    | cell_int_option
    """
    p[0] = p[1]


def p_cell_int_option(p):
    """cell_int_option : U integer
    | MAT integer
    | LAT integer
    """
    p[0] = {p[1]: p[2]}


def p_cell_float_option(p):
    """cell_float_option : IMP ':' particle_list float
    | TMP float
    | RHO float
    | VOL float
    """
    if len(p) == 5:
        p[0] = {p[1] + par: p[4] for par in p[3]}
    else:
        p[0] = {p[1]: p[2]}


def p_trcl_option(p):
    """trcl_option : '*' TRCL '(' transform_params ')'
    | TRCL '(' transform_params ')'
    | TRCL int_number
    """
    if len(p) == 3:
        p[0] = {"TRCL": p[2]}
    elif len(p) == 5:
        p[0] = {"TRCL": p[3]}
    else:
        p[4]["indegrees"] = True
        p[0] = {"TRCL": p[4]}


def p_fill_option(p):
    """fill_option : '*' FILL int_number '(' transform_params ')'
    | FILL int_number '(' transform_params ')'
    | FILL int_number '(' int_number ')'
    | FILL int_number
    """
    if len(p) == 7:
        p[5]["indegrees"] = True
        p[0] = {"FILL": {"universe": p[3], "transform": p[5]}}
    elif len(p) == 6:
        p[0] = {"FILL": {"universe": p[2], "transform": p[4]}}
    else:
        p[0] = {"FILL": {"universe": p[2]}}


def p_cell_material(p):
    """cell_material : int_number float
    | void_material
    """
    if len(p) == 2:
        p[0] = None
    else:
        p[0] = p[1], p[2]


def p_expression(p):
    """expression : expression ':' term
    | term
    """
    if len(p) == 4:
        p[0] = p[1] + p[3]
        p[0].append("U")
    else:
        p[0] = p[1]


def p_term(p):
    """term : term factor
    | factor
    """
    if len(p) == 3:
        p[0] = p[1] + p[2]
        p[0].append("I")
    else:
        p[0] = p[1]


def p_factor(p):
    """factor : '#' '(' expression ')'
    | '(' expression ')'
    | '-' int_number
    | '+' int_number
    | '#' int_number
    | int_number
    """
    if len(p) == 2:
        p[0] = [p[1]]
    elif len(p) == 3:
        p[0] = [p[2]]
        if p[1] == "-":
            p[0].append("C")
        elif p[1] == "#":
            p[0].append("#")
    elif len(p) == 4:
        p[0] = p[2]
    elif len(p) == 5:
        p[0] = p[3] + ["C"]


def p_surface_cards(p):
    """surface_cards : surface_cards separator surface_card
    | surface_card
    """
    if len(p) == 2:
        p[0] = {p[1]["name"]: p[1]}
    elif len(p) == 4:
        name = p[3]["name"]
        if name in p[1].keys():
            raise ValueError("Duplicate surface name {0}".format(name))
        p[1][name] = p[3]
        p[0] = p[1]


def p_surface_card(p):
    """surface_card : '*' int_number surface_description
    | '+' int_number surface_description
    | int_number surface_description
    """
    n = len(p)
    surf = p[n - 1]
    surf["name"] = p[n - 2]
    if n == 4:
        surf["modifier"] = p[1]
    comment = extract_comments(p.lineno(1))
    if comment:
        surf["comment"] = comment
    p[0] = surf


def p_surface_description(p):
    """surface_description : integer surface_type param_list
    | surface_type param_list"""
    n = len(p)
    descr = {"params": p[n - 1], "kind": p[n - 2]}
    if n == 4:
        descr["transform"] = p[1]
    p[0] = descr


def p_param_list(p):
    """param_list : param_list float
    | float
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[2]]


def p_float(p):
    """float : '+' flt_number
    | '-' flt_number
    | flt_number
    | integer
    """
    if p[1] == "-":
        p[0] = -p[2]
    elif p[1] == "+":
        p[0] = p[2]
    else:
        p[0] = p[1]


def p_integer(p):
    """integer : '+' int_number
    | '-' int_number
    | int_number
    """
    if p[1] == "-":
        p[0] = -p[2]
    elif p[1] == "+":
        p[0] = p[2]
    else:
        p[0] = p[1]


def p_data_cards(p):
    """data_cards : data_cards separator data_card
    | data_card
    """
    if len(p) == 2:
        name, value = p[1]
        p[0] = {name: value}
    elif len(p) == 4:
        name, value = p[3]
        if name in p[1].keys() and isinstance(value, dict):
            p[1][name].update(value)
        else:
            p[1][name] = value
        p[0] = p[1]


def p_data_card(p):
    """data_card : mode_card
    | material_card
    | transform_card
    """
    p[0] = p[1]


def p_mode_card(p):
    """mode_card : MODE particle_list"""
    p[0] = "MODE", p[2]


def p_particle_list(p):
    """particle_list : particle_list particle
    | particle
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[2]]


def p_particle(p):
    """particle : N
    | P
    | E
    """
    p[0] = p[1]


def p_transform_card(p):
    """transform_card : '*' TR int_number transform_params
    | TR int_number transform_params
    """
    n = len(p)
    tr = p[n - 1]
    tr["name"] = p[n - 2]
    if n == 5:
        tr["indegrees"] = True
    comment = extract_comments(p.lineno(1))
    if comment:
        tr["comment"] = comment
    p[0] = "TR", {tr["name"]: tr}


def p_transform_params(p):
    """transform_params : translation rotation integer
    | translation rotation
    | translation
    """
    tr = {"translation": p[1]}
    if len(p) > 2:
        tr["rotation"] = p[2]
    if len(p) == 4:
        tr["inverted"] = True
    p[0] = tr


def p_translation(p):
    """translation : float float float"""
    p[0] = [p[1], p[2], p[3]]


def p_rotation(p):
    """rotation : float float float float float float float float float
    | float float float float float float
    | float float float float float
    | float float float
    """
    p[0] = [p[i] for i in range(1, len(p))]


def p_material_card(p):
    """material_card : M int_number composition_list material_options
    | M int_number composition_list
    """
    if len(p) == 5:
        p[3].update(p[4])
    p[3]["name"] = p[2]
    comment = extract_comments(p.lineno(1))
    if comment:
        p[3]["comment"] = comment
    p[0] = "M", {p[2]: p[3]}


def p_composition_list(p):
    """composition_list : composition_list zaid_fraction
    | zaid_fraction
    """
    if len(p) == 2:
        key = p[1][0]
        value = p[1][1:]
        p[0] = {key: [value]}
    else:
        key = p[2][0]
        value = p[2][1:]
        if key in p[1].keys():
            p[1][key].append(value)
        else:
            p[1][key] = [value]
        p[0] = p[1]


def p_zaid_fraction(p):
    """zaid_fraction : int_number '.' lib_spec float
    | int_number float
    """
    n = len(p)
    key = "atomic" if p[n - 1] > 0 else "weight"
    opts = {}
    if n == 5:
        opts["lib"] = p[3]
    p[0] = key, p[1], opts, abs(p[n - 1])


def p_material_options(p):
    """material_options : material_options material_option
    | material_option
    """
    if len(p) == 2:
        p[0] = {p[1][0]: p[1][1]}
    else:
        p[1][p[2][0]] = p[2][1]
        p[0] = p[1]


def p_material_option(p):
    """material_option : GAS integer
    | ESTEP integer
    | NLIB lib_spec
    | PLIB lib_spec
    | PNLIB lib_spec
    | ELIB lib_spec
    | COND integer
    """
    p[0] = p[1], p[2]


mcnp_input_parser = yacc.yacc(
    tabmodule="mcnp_input_tab",
    debug=__DEBUG__,
    debuglog=logger,
    errorlog=logger if __DEBUG__ else yacc.NullLogger(),
)


def read_mcnp(filename: Union[str, Path], encoding: str = "cp1251") -> Universe:
    warn(
        "The function 'read_mcnp' is deprecated. Use mckit.parser.from_file() instead.",
        DeprecationWarning,
    )
    with open(filename, encoding=encoding) as f:
        text = f.read()
    return read_mcnp_text(text)


def read_mcnp_text(
    text: str,
) -> Universe:
    warn(
        "The function 'read_mcnp_text' is deprecated. Use mckit.parser.from_text() instead.",
        DeprecationWarning,
    )
    text = (
        text.rstrip() + "\n"
    )  # float number spec requires end of line or space after float
    mcnp_input_lexer.begin("INITIAL")
    title, cells, surfaces, data = mcnp_input_parser.parse(
        text, tracking=True, lexer=mcnp_input_lexer, debug=__DEBUG__
    )
    bodies = []
    for name in list(cells.keys()):
        bodies.append(_get_cell(name, cells, surfaces, data))
    return produce_universes(bodies)


def _get_transformation(name, data):
    if isinstance(name, dict):
        return Transformation(**name)
    assert isinstance(name, int)
    tr = data["TR"][name]
    if isinstance(tr, dict):
        tr = Transformation(**tr)
        data["TR"][name] = tr
    return tr


def _get_composition(name, data):
    comp = data["M"][name]
    if isinstance(comp, dict):
        for i, (el, opt, frac) in enumerate(comp.get("atomic", [])):
            comp["atomic"][i] = (Element(el, **opt), frac)
        for i, (el, opt, frac) in enumerate(comp.get("weight", [])):
            comp["weight"][i] = (Element(el, **opt), frac)
        comp = Composition(**comp)
        data["M"][name] = comp
    return comp


def _get_surface(name, surfaces, data):
    surf_data = surfaces[name]
    if isinstance(surf_data, dict):
        kind = surf_data.pop("kind")
        params = surf_data.pop("params")
        tr = surf_data.get("transform", None)
        if isinstance(tr, int):
            surf_data["transform"] = _get_transformation(tr, data)
        surf = create_surface(kind, *params, **surf_data)
        surfaces[name] = surf
    else:
        surf = surf_data
    return surf


def _get_cell(name, cells, surfaces, data):
    if isinstance(cells[name], Body):  # Already exists
        return cells[name]
    cell_data = cells.pop(name)  # To avoid circular references and infinite
    # cycle
    geometry = cell_data.pop("geometry", None)
    if geometry is None:  # find reference cell
        ref_name = cell_data.pop("reference")
        ref_cell = _get_cell(ref_name, cells, surfaces, data)
        geometry = ref_cell.shape
        for k, v in ref_cell.options.items():
            if k not in cell_data.keys():
                cell_data[k] = v
        rho = cell_data.pop("RHO", None)
        material = cell_data["MAT"]
        cell_data["MAT"] = {"composition": material.composition.name()}
        if rho is not None:
            if rho > 0:
                cell_data["MAT"]["concentration"] = rho * 1.0e24
            else:
                cell_data["MAT"]["density"] = abs(rho)
        else:
            cell_data["MAT"]["density"] = material.density
    else:  # create geometry from polish notation
        for i, g in enumerate(geometry):
            if isinstance(g, int):
                if i + 1 < len(geometry) and geometry[i + 1] == "#":
                    comp_cell = _get_cell(g, cells, surfaces, data)
                    shape = comp_cell.shape
                    tr = comp_cell.options.get("TRCL", None)
                    if tr:
                        shape = shape.transform(tr)
                    geometry[i] = shape
                    geometry[i + 1] = "C"
                else:
                    geometry[i] = _get_surface(g, surfaces, data)

    # Create material if necessary
    mat_data = cell_data.get("MAT", None)
    if mat_data:
        comp = mat_data["composition"]
        mat_data["composition"] = _get_composition(comp, data)
        cell_data["MAT"] = Material(**mat_data)

    # Replace transformations
    if "TRCL" in cell_data.keys():
        cell_data["TRCL"] = _get_transformation(cell_data["TRCL"], data)
    fill = cell_data.get("FILL", {})
    temp_tr = None
    if "transform" in fill.keys():
        temp_tr = fill["transform"]
        temp_tr = _get_transformation(temp_tr, data)
        if temp_tr and not isinstance(temp_tr, Transformation):
            print(name)
        fill["transform"] = temp_tr

    cell = Body(geometry, **cell_data)
    cells[name] = cell
    return cell



# -*- coding: utf-8 -*-
"""
Сборка модели из конвертов и входяших в них юниверсов по заданной спецификации.
"""
from typing import Dict, List, Optional, Tuple, Union

from functools import reduce
from pathlib import Path

import mckit as mk
import numpy as np
import tomlkit as tk

from mckit import Transformation
from mckit.cli.logging import logger
from mckit.parser.mcnp_input_sly_parser import ParseResult, from_file
from mckit.utils import filter_dict
from tomlkit import items as tk_items

from .common import save_mcnp


def compose(output, fill_descriptor_path, source, override):
    logger.info("Loading model from {s}", s=source)
    parse_result: ParseResult = from_file(source)
    envelopes = parse_result.universe
    source = Path(source)
    universes_dir = source.absolute().parent
    assert universes_dir.is_dir()
    logger.info("Loading fill-descriptor from {f}", f=fill_descriptor_path)
    with fill_descriptor_path.open() as fid:
        fill_descriptor = tk.parse(fid.read())

    universes = load_universes(fill_descriptor, universes_dir)
    named_transformations = load_named_transformations(fill_descriptor)

    comps = {}

    for k, v in universes.items():
        u, _ = v
        cps = u.get_compositions()
        comps[k] = {c for c in cps}

    common = reduce(set.union, comps.values())
    envelopes.set_common_materials(common)
    cells_index = dict((cell.name(), cell) for cell in envelopes)

    for i, spec in universes.items():
        universe, transformation = spec
        universe.set_common_materials(common)
        cell = cells_index[i]
        cell.options = filter_dict(cell.options, "original")
        cell.options["FILL"] = {"universe": universe}
        if transformation is not None:
            if isinstance(transformation, tk_items.Array):
                transformation1 = np.fromiter(
                    map(float, iter(transformation)), dtype=np.double
                )
                try:
                    translation = transformation1[:3]
                    if len(transformation1) > 3:
                        rotation = transformation1[3:]
                    else:
                        rotation = None
                    transformation2 = mk.Transformation(
                        translation=translation,
                        rotation=rotation,
                        indegrees=True,  # Assuming that on decompose we store a transformation in degrees as well
                    )
                except ValueError as ex:
                    raise ValueError(
                        f"Failed to process FILL transformation in cell #{cell.name()} of universe #{universe.name()}"
                    ) from ex
                cell.options["FILL"]["transform"] = transformation2
            elif isinstance(transformation, tk_items.Integer):
                assert (
                    named_transformations is not None
                ), "There are no named transformations in the fill descriptor file"
                transformation1 = named_transformations[int(transformation)]
                cell.options["FILL"]["transform"] = transformation1
            else:
                raise NotImplementedError(
                    f"Unexpected type of transformation parameter {type(transformation)}"
                )
    save_mcnp(envelopes, output, override)


def load_universes(
    fill_descriptor, universes_dir
) -> Dict[int, Tuple[mk.Universe, Union[int, List[float]]]]:
    filler_path_map: Dict[int, Tuple[Path, mk.Universe]] = dict()
    cell_filler_map: Dict[int, Tuple[mk.Universe, Union[int, List[float]]]] = dict()

    for k, v in fill_descriptor.items():
        if isinstance(v, dict) and "universe" in v:
            cell_name = int(k)
            universe_name = int(v["universe"])
            transformation = v.get("transform", None)
            universe_path = Path(v["file"])
            if universe_name in filler_path_map:
                prev_path, prev_universe = filler_path_map[universe_name]
                if prev_path == universe_path:
                    universe = prev_universe
                else:
                    raise ValueError(
                        f'\n \
                        Filler number {universe_name} specifies different paths to load \n \
                        "{prev_path}" and "{universe_path}"'
                    )
            else:
                if universe_path.exists():
                    load_path = universe_path
                else:
                    load_path = universes_dir / universe_path
                if not load_path.exists():
                    raise FileNotFoundError(universe_path)
                logger.info("Loading file {u}", u=load_path)
                parse_result: ParseResult = from_file(load_path)
                universe = parse_result.universe
                universe.rename(name=universe_name)
                filler_path_map[universe_name] = (universe_path, universe)

            cell_filler_map[cell_name] = (universe, transformation)

    return cell_filler_map


def load_named_transformations(fill_descriptor) -> Optional[Dict[int, Transformation]]:
    transformations = fill_descriptor.get("named_transformations", None)
    if transformations:
        named_transformations = {}
        for k, v in transformations.items():
            name = int(k[2:])
            transform_params = np.fromiter(map(float, v), dtype=float)
            translation = transform_params[:3]
            if transform_params.size == 9:
                rotation = transform_params[3:]
            else:
                rotation = None
            transform = Transformation(
                translation=translation, rotation=rotation, indegrees=True, name=name
            )
            named_transformations[name] = transform
        return named_transformations
    else:
        return None



import re

import ply.lex as lex
import ply.yacc as yacc

from mckit.material import Element

# TODO dvp: the TIME_UNITS should be IntEnum and change 365*24 to (365*24+6)
TIME_UNITS = {
    "SECS": 1.0,
    "MINS": 60.0,
    "HOURS": 3600.0,
    "DAYS": 3600.0 * 24,
    "YEARS": 3600.0 * 24 * 365,
}

TIME_ALIAS = {
    "s": TIME_UNITS["SECS"],
    "m": TIME_UNITS["MINS"],
    "h": TIME_UNITS["HOURS"],
    "d": TIME_UNITS["DAYS"],
    "y": TIME_UNITS["YEARS"],
    "ky": TIME_UNITS["YEARS"] * 1000,
}

literals = ["+", "-", "(", ")"]


# List of token names
tokens = ["newline", "int_number", "flt_number", "keyword", "TIME", "INTERVAL"]


NEWLINE = r"\n"
EXPONENT = r"(E[-+]?\d+)"
INT_NUMBER = r"(\d+)"
FLT_NUMBER = (
    INT_NUMBER
    + r"?"
    + r"\."
    + INT_NUMBER
    + EXPONENT
    + r"?|"
    + INT_NUMBER
    + r"\."
    + r"?"
    + EXPONENT
    + r"|"
    + INT_NUMBER
    + r"\."
)
KEYWORD = r"[A-Z]+(/[A-Z]+)?"
SKIP = r"[. ]"

t_ignore = SKIP


@lex.TOKEN(NEWLINE)
def t_newline(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    return t


@lex.TOKEN(FLT_NUMBER)
def t_flt_number(t):
    t.value = float(t.value)
    return t


@lex.TOKEN(INT_NUMBER)
def t_int_number(t):
    t.value = int(t.value)
    return t


@lex.TOKEN(KEYWORD)
def t_keyword(t):
    value = t.value.upper()
    if value == "TIME" or value == "INTERVAL":
        t.type = value
        t.value = value
    return t


def t_error(t):
    column = t.lexer.lexpos - t.lexer.last_pos + 1
    msg = r"Illegal character '{0}' at line {1} column {2}".format(
        t.value[0], t.lexer.lineno, column
    )
    raise ValueError(msg, t.value[0], t.lexer.lineno, column)


fispact_lexer = lex.lex(reflags=re.MULTILINE + re.IGNORECASE + re.VERBOSE)


def p_data(p):
    """data : data timeframe
    | timeframe
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[1].append(p[2])
        p[0] = p[1]


def p_timeframe(p):
    """timeframe : timeheader isotope_data
    | timeheader gamma_data
    | timeheader
    """
    index, interval, time = p[1]
    frame = {"index": index, "duration": interval, "time": time}
    if len(p) == 3:
        frame.update(p[2])
    p[0] = frame


def p_timeheader(p):
    """timeheader : TIME flt_number keyword INTERVAL int_number INTERVAL TIME flt_number keyword newline"""
    index = p[5]
    time = p[2] * TIME_ALIAS[p[3]]
    interval = p[8] * TIME_UNITS[p[9]]
    p[0] = index, interval, time


def p_isotope_data(p):
    """isotope_data : isotope_data isotope_row newline
    | isotope_row newline
    """
    n = len(p)
    elem, data1, data2 = p[n - 2]
    if n == 3:
        p[0] = {"data1": {elem: data1}, "data2": {elem: data2}}
    else:
        p[1]["data1"][elem] = data1
        p[1]["data2"][elem] = data2
        p[0] = p[1]


def p_isotope_row(p):
    """isotope_row : isotope flt_number flt_number
    | isotope int_number flt_number flt_number
    """
    n = len(p)
    p[0] = p[1], p[n - 2], p[n - 1]


def p_isotope(p):
    """isotope : keyword int_number
    | keyword int_number keyword
    """
    name = p[1] + str(p[2])
    if len(p) == 4:
        isomer = ord(p[3]) - ord("m") + 1
    else:
        isomer = 0
    p[0] = Element(name, isomer=isomer)


def p_gamma_data(p):
    """gamma_data : fiss ab ab tot_gamma spectrum"""
    p[5]["fissions"] = p[1]
    p[5]["a-energy"] = p[2]
    p[5]["b-energy"] = p[3]
    p[5]["g-energy"] = p[4]
    p[0] = p[5]


def p_fiss(p):
    """fiss : keyword keyword keyword flt_number newline"""
    p[0] = p[4]


def p_ab(p):
    """ab : keyword '-' keyword keyword flt_number newline"""
    p[0] = p[5]


def p_tot_gamma(p):
    """tot_gamma : keyword keyword keyword flt_number newline"""
    p[0] = p[4]


def p_spectrum(p):
    """spectrum : spectrum bin newline
    | bin newline
    """
    n = len(p)
    low, high, data1, data2 = p[n - 2]
    if n == 3:
        p[0] = {"ebins": [low, high], "data1": [data1], "data2": [data2]}
    else:
        p[1]["ebins"].append(high)
        p[1]["data1"].append(data1)
        p[1]["data2"].append(data2)
        p[0] = p[1]


def p_bin(p):
    """bin : '(' flt_number '-' flt_number keyword ')' flt_number flt_number"""
    p[0] = p[2], p[4], p[7], p[8]


def p_error(p):
    if p:
        column = p.lexer.lexpos - p.lexer.last_pos + 1
        print(
            "Syntax error at token {0} {3}, line {1}, column {2}".format(
                p.type, p.lexer.lineno, column, p.value
            )
        )
    else:
        print("Syntax error at EOF")


fispact_parser = yacc.yacc(tabmodule="fispact_tab", debug=True)


def read_fispact_tab(filename):
    """Reads FISPACT output tab file.

    This function is common for all tab files (1-4). Output data format depends
    on file extension: tab1 - number of atoms of each nuclide, tab2 - activity
    of each nuclide, tab3 - ingestion and inhalation doses,
    tab4 - gamma-radiation spectrum.

    The result is a list of time frames. Every time frame is a dictionary, that
    contains the following data:
        * index - the number of frame. Numbers start from 1;
        * time - total time passed from the start of irradiation or relaxation
                 in seconds;
        * duration - time passed from the previous frame also in seconds;
        * atoms - dictionary[Element -> float] - the number of atoms for each
                 element. Only for tab1 file;
        * activity - dictionary[Element -> float] - the activity of each
                 element in Bq. Only for tab2 file;
        * ingestion - dictionary[Element -> float] - ingestion dose for every
                 isotope in Sv/h. Only for tab3 file;
        * inhalation - dictionary[Element -> float] - inhalation dose for every
                 isotope in Sv/h. Only for tab3 file;
        * fissions - the number of spontaneous fissions. Only for tab4 file;
        * a-energy - energy of alpha radiation in MeV/sec. Only for tab4 file.
        * b-energy - energy of beta radiation in MeV/sec. Only for tab4 file.
        * g-energy - energy of gamma radiation in MeV/sec. Only for tab4 file.
        * ebins - energy bins for gamma-radiation in MeV. Only for tab4 file.
        * flux - gamma-radiation group flux in photons per cc per sec. Only for
                 tab4 file.

    Parameters
    ----------
    filename : str
        Name of FISPACT output tab file.

    Returns
    -------
    time_frames : list
        List of time frames.
    """
    ext = filename.rpartition(".")[2].lower()
    with open(filename) as f:
        text = f.read()
    fispact_lexer.lineno = 1
    fispact_lexer.linepos = 1
    fispact_lexer.last_pos = 1
    time_frames = fispact_parser.parse(text, lexer=fispact_lexer)
    time = 0
    for tf in time_frames:
        time += tf["duration"]
        tf["time"] = time
        data1 = tf.pop("data1", None)
        data2 = tf.pop("data2", None)
        if data1 is None and data2 is None:
            continue
        if ext == "tab1":
            tf["atoms"] = data1
        elif ext == "tab2":
            tf["activity"] = data1
        elif ext == "tab3":
            tf["ingestion"] = data1
            tf["inhalation"] = data2
        elif ext == "tab4":
            tf["flux"] = data2
    return time_frames



from typing import Any, Dict, Tuple, cast

import collections
import collections.abc as abc
import functools
import itertools

from copy import deepcopy

import numpy as np

from numpy import ndarray

from ..constants import FLOAT_TOLERANCE

MAX_DIGITS = np.finfo(float).precision

# def digits_in_fraction_for_str(
#         value: float,
#         reltol: float = FLOAT_TOLERANCE,
#         resolution: float = None
# ) -> int:
#     if value == 0.0:
#         return 0
#     if value < 0.0:
#         value = -value
#     if resolution and value < resolution:
#         return 0
#     max_remainder = value * reltol
#     s_value = str(value)
#     _, s_rem = s_value.split('.')
#
#     def _iter():
#         ord0 = ord('0')
#         m = 0.1
#         for c in s_rem:
#             yield m * (ord(c) - ord0)
#             m *= 0.1
#
#     rem = np.flip(np.fromiter(_iter(), float))
#     n = np.searchsorted(rem, max_remainder)
#     return rem.size - n


def significant_digits(value, reltol=FLOAT_TOLERANCE, resolution=None):
    """The minimum number of significant digits to provide relative tolerance.

    Parameters
    ----------
    value : float
        The value to be checked.
    reltol : float
        Relative tolerance needed to represent the value.
    resolution : float
        The threshold value, below which numbers are believed to be zero.
        Default: None - not applied.

    Returns
    -------
    digits : int
        The number of digits.
    """
    if value == 0.0 or resolution and abs(value) < resolution:
        return 0
    dec = get_decades(value)
    low = min(dec, 0)
    high = MAX_DIGITS
    d = d0 = abs(value)
    while high - low > 1:
        p = round(0.5 * (high + low))
        v = round(value, p)
        d = max(d0, abs(v))
        if abs(value - v) > reltol * d:
            low = p
        else:
            high = p
    v = round(value, low)
    if abs(value - v) < reltol * d:
        return low
    else:
        return high


# LG2 = math.log10(2.0)


def get_decades(value):
    # TODO dvp: check if math.frexp is applicable, this mostly works but some test for pretty_print fail.
    # if value == 0.0:
    #     return 0
    # mantissa, exponent = math.frexp(value)  # type: float, int
    # if -0.5 <= mantissa <= 0.5:
    #     decades: int = math.floor(LG2 * (exponent - 1))
    # else:
    #     decades: int = math.floor(LG2 * exponent)
    # return decades

    if value != 0:
        decimal_power = np.log10(
            abs(value)
        )  # TODO dvp: log10 will be called billion times on C-model
    else:
        decimal_power = 0
    decades = np.trunc(decimal_power)
    if decimal_power < 0:
        decades -= 1
    return int(decades)


def significant_array(
    array: ndarray, reltol: float = FLOAT_TOLERANCE, resolution: float = None
) -> ndarray:
    """The minimum number of significant digits to provide the desired relative and absolute tolerances."""
    result: ndarray = np.empty_like(array, dtype=int)
    for index in zip(*map(np.ravel, np.indices(array.shape))):
        result[index] = significant_digits(array[index], reltol, resolution)
    return result


def round_scalar(value, digits=None):
    """Rounds scalar value to represent the value in minimal form.

    Parameters
    ----------
    value : float
        The value to be rounded.
    digits : int
        The number of significant digits.

    Returns
    -------
    result : float
        Rounded value.
    """
    if digits is None:
        digits = significant_digits(value, FLOAT_TOLERANCE, FLOAT_TOLERANCE)
    return round(value, digits)


def round_array(array: ndarray, digits_array: ndarray = None) -> ndarray:
    """Rounds array to desired precision.

    Parameters
    ----------
    array :
        Array of values.
    digits_array :
        Array of corresponding significant digits.

    Returns
    -------
    result :
        Rounded array.
    """
    if digits_array is None:
        digits_array = significant_array(array, FLOAT_TOLERANCE, FLOAT_TOLERANCE)
    result: ndarray = np.empty_like(array)
    for index in zip(*map(np.ravel, np.indices(array.shape))):
        result[index] = round_scalar(array[index], digits_array[index])
    return result


@functools.singledispatch
def are_equal(a, b) -> bool:
    return a is b or a == b


@are_equal.register
def _(a: str, b: str) -> bool:
    return a is b or a == b


@are_equal.register
def _(a: ndarray, b: ndarray) -> bool:
    return np.array_equal(a, b)


@are_equal.register
def _(a: collections.abc.Iterable, b) -> bool:
    if not issubclass(type(b), collections.abc.Iterable):
        return False
    for ai, bi in itertools.zip_longest(a, b):
        if not are_equal(ai, bi):
            return False
    return True


@functools.singledispatch
def is_in(where, x) -> bool:
    if where is None:
        return False
    return x is where or x == where


@is_in.register
def _(where: str, x) -> bool:
    return x is where or x == where


@is_in.register
def _(where: tuple, x) -> bool:
    for i in where:
        if is_in(i, x):
            return True
    return False


@is_in.register
def _(where: collections.abc.Callable, x) -> bool:
    return where(x)  # type: ignore


@is_in.register
def _(where: collections.abc.Container, x) -> bool:
    return x in where


def filter_dict(a: Dict[Any, Any], *drop_items) -> Dict[Any, Any]:
    res = {}
    for k, v in a.items():
        # if drop_items is None or not check_if_is_in(k, *drop_items):
        if drop_items and is_in(drop_items, k):
            pass
        else:
            if isinstance(v, dict):
                res[k] = filter_dict(v, *drop_items)
            elif issubclass(type(v), collections.abc.Collection):
                res[k] = deepcopy(v)
            else:
                res[k] = v
    return res


@functools.singledispatch
def make_hashable(x):
    raise TypeError(f"Don't know how to make {type(x).__name__} objects hashable")


@make_hashable.register
def _(x: collections.abc.Hashable):
    return x


@make_hashable.register
def _(x: str):
    return x


@make_hashable.register
def _(x: collections.abc.Mapping) -> Tuple:
    return tuple(map(lambda i: (i[0], make_hashable(i[1])), x.items()))


@make_hashable.register
def _(x: collections.abc.Iterable) -> Tuple:
    return tuple(map(make_hashable, x))


def make_hash(*items) -> int:
    if 1 < len(items):
        return make_hash(tuple(map(make_hash, items)))
    return hash(make_hashable(items[0]))


def is_sorted(a: np.ndarray) -> bool:
    return cast(bool, np.all(np.diff(a) > 0))


def mids(a: np.ndarray) -> np.ndarray:
    result: ndarray = 0.5 * (a[1:] + a[:-1])
    return result


def prettify_float(x: float, fmt: str = "{:.13g}") -> str:
    if x.is_integer():
        return str(int(x))
    else:
        return fmt.format(x)
